import multiprocessing as mp
import os
from copy import deepcopy
from difflib import SequenceMatcher
from itertools import permutations

import chardet
from pygments import lexers
from pygments.util import ClassNotFound

from tlv.cls_ast import TLVAst
from tlv.cls_ast import TLVFinding
from tlv.helper import debug
from tlv.helper import die


def parse(args):
    res = []
    for f in args.files:
        if not os.path.isfile(f):
            continue
        _lexer = None
        if args.lexer:
            try:
                _lexer = lexers.get_lexer_by_name(args.lexer)
            except ClassNotFound:
                die(-1, "Lexer '{lex}' isn't available can't proceed".format(lex=args.lexer))
        else:
            try:
                _lexer = lexers.get_lexer_for_filename(f)
            except ClassNotFound:
                continue
        debug(args, "Reading {f}".format(f=f))
        with open(f, "rb") as i:
            try:
                content = i.read()
                _enc = chardet.detect(content)
                content = content.decode(_enc["encoding"]).encode("utf-8")
                _ast = TLVAst(f)
                res.append(_ast)
                for x in _lexer.get_tokens(content):
                    _type, _value = x
                    if str(_type) in args.wildcard_token:
                        _ast.Processed.append((_type, ""))
                    else:
                        _ast.Processed.append((_type, _value))
                    _ast.Raw.append(x)
            except UnicodeDecodeError:
                continue
    return res


def __evaluate(args, first, second):
    findings = []
    debug(args, "Comparing {a}<>{b}".format(a=os.path.basename(
        first.Filename), b=os.path.basename(second.Filename)))
    m = SequenceMatcher(None, first.Processed,
                        second.Processed).get_matching_blocks()
    for _m in m:
        _id = TLVFinding.Validate(args, first, second, _m)
        if _id:
            findings.append(TLVFinding(first, second, _m, _id))
            findings.append(TLVFinding(second, first, _m, _id))
    return findings


def validate(args, _list):
    perms = permutations(_list, 2)
    results = []
    with mp.Pool(processes=args.jobs) as pool:
        results = [pool.apply(__evaluate, args=(args, first, second))
                   for first, second in perms]
    # append one time every file compared against it self
    # by cutting the content into half
    for x in _list:
        _x1 = deepcopy(x)
        _x2 = deepcopy(x)
        chop_point = int(len(_x1.Raw) / 2)
        if chop_point > 0:
            _line_offset = TLVFinding.get_text(
                _x1, 0, chop_point).count(os.linesep)
            _x1.Raw = _x1.Raw[0: chop_point]
            _x1.Processed = _x1.Processed[0: chop_point]
            _x2.Raw = _x2.Raw[chop_point:]
            _x2.Processed = _x2.Processed[chop_point:]
            _x2.LineOffset = _line_offset
            results.append(__evaluate(args, _x1, _x2))
    return list(set([item for sublist in results for item in sublist]))  # noqa: C403



from tlv.arg_parser import ArgParser
from tlv.src_parser import parse
from tlv.src_parser import validate


def main():
    args = ArgParser()
    res = validate(args, parse(args))
    for r in res:
        r.Print(args)


if __name__ == '__main__':
    main()



LEXERS = [
    "3d",
    "aap",
    "actionscript",
    "actionscript3",
    "ada",
    "ada2005",
    "ada95",
    "adl",
    "agda",
    "ahk",
    "al",
    "alloy",
    "amienttalk",
    "amienttalk/2",
    "ampl",
    "an",
    "antlr",
    "antlr-actionscript",
    "antlr-as",
    "antlr-c#",
    "antlr-cpp",
    "antlr-csharp",
    "antlr-java",
    "antlr-perl",
    "antlr-python",
    "antlr-r",
    "antlr-ruy",
    "apachecon",
    "apl",
    "applescript",
    "arduino",
    "arexx",
    "as",
    "as3",
    "asemake",
    "asic",
    "asm",
    "aspectj",
    "aspx-cs",
    "asy",
    "asymptote",
    "at",
    "atch",
    "augeas",
    "autohotkey",
    "autoit",
    "awk",
    "c",
    "c#",
    "c++",
    "c++-ojdum",
    "c-ojdump",
    "ca65",
    "cadl",
    "camkes",
    "capdl",
    "capnp",
    "casic",
    "ceylon",
    "chai",
    "chaiscript",
    "chapel",
    "charmci",
    "cheetah",
    "chpl",
    "cirru",
    "cl",
    "clay",
    "clean",
    "clj",
    "cljs",
    "clojure",
    "clojurescript",
    "cmake",
    "cmas",
    "co",
    "code",
    "common-lisp",
    "componentpascal",
    "console",
    "control",
    "cool",
    "coq",
    "cp",
    "cpp",
    "cpp-ojdump",
    "cpsa",
    "cr",
    "crmsh",
    "croc",
    "cry",
    "cryptol",
    "crystal",
    "csh",
    "csharp",
    "csound",
    "csound-csd",
    "csound-document",
    "csound-orc",
    "csound-sco",
    "csound-score",
    "css",
    "css+django",
    "css+er",
    "css+genshi",
    "css+genshitext",
    "css+jinja",
    "css+lasso",
    "css+mako",
    "css+mozpreproc",
    "css+php",
    "css+ruy",
    "css+smarty",
    "cu",
    "cucumer",
    "cuda",
    "cxx-ojdump",
    "cypher",
    "cython",
    "d",
    "d-ojdump",
    "dart",
    "dasm16",
    "decontrol",
    "delphi",
    "desources",
    "dg",
    "di",
    "django",
    "do",
    "docker",
    "dosatch",
    "doscon",
    "dpatch",
    "dtd",
    "duy",
    "dylan",
    "dylan-console",
    "dylan-lid",
    "dylan-repl",
    "e",
    "earl-grey",
    "earlgrey",
    "easytrieve",
    "ec",
    "ecl",
    "eg",
    "ei",
    "elisp",
    "elixir",
    "elm",
    "emacs",
    "emacs-lisp",
    "en",
    "er",
    "erl",
    "erlang",
    "evoque",
    "ex",
    "exs",
    "extempore",
    "ezhil",
    "gap",
    "gas",
    "gawk",
    "genshi",
    "genshitext",
    "gherkin",
    "glsl",
    "gnuplot",
    "go",
    "golo",
    "gooddata-cl",
    "gosu",
    "gro",
    "groovy",
    "gst",
    "haml",
    "handlears",
    "haskell",
    "haxe",
    "haxeml",
    "hexdump",
    "hlsl",
    "hs",
    "hsa",
    "hsail",
    "hspec",
    "html",
    "html+cheetah",
    "html+django",
    "html+er",
    "html+evoque",
    "html+genshi",
    "html+handlears",
    "html+jinja",
    "html+kid",
    "html+lasso",
    "html+mako",
    "html+myghty",
    "html+ng2",
    "html+php",
    "html+ruy",
    "html+smarty",
    "html+spit",
    "html+twig",
    "htmldjango",
    "http",
    "hx",
    "hxml",
    "hxsl",
    "hylang",
    "i",
    "i6t",
    "idl",
    "idl4",
    "idr",
    "idris",
    "iex",
    "igor",
    "igorpro",
    "ik",
    "in",
    "ini",
    "io",
    "ioke",
    "ir",
    "irc",
    "isaelle",
    "itex",
    "j",
    "jade",
    "jags",
    "jasmin",
    "jasminxt",
    "java",
    "javascript",
    "javascript+cheetah",
    "javascript+django",
    "javascript+er",
    "javascript+genshi",
    "javascript+genshitext",
    "javascript+jinja",
    "javascript+lasso",
    "javascript+mako",
    "javascript+myghty",
    "javascript+php",
    "javascript+ruy",
    "javascript+smarty",
    "jcl",
    "jinja",
    "jl",
    "jlcon",
    "jproperties",
    "js",
    "js+cheetah",
    "js+django",
    "js+er",
    "js+genshi",
    "js+genshitext",
    "js+jinja",
    "js+lasso",
    "js+mako",
    "js+myghty",
    "js+php",
    "js+ruy",
    "js+smarty",
    "js+spit",
    "jsg",
    "json",
    "json-ld",
    "json-oject",
    "jsonld",
    "jsp",
    "julia",
    "juttle",
    "kal",
    "kcon",
    "kid",
    "koka",
    "kotlin",
    "lagda",
    "lasso",
    "lassoscript",
    "lcry",
    "lcryptol",
    "lean",
    "less",
    "lhaskell",
    "lhs",
    "lid",
    "lidr",
    "lidris",
    "lighttpd",
    "lighty",
    "limo",
    "liquid",
    "lisp",
    "literate-agda",
    "literate-cryptol",
    "literate-haskell",
    "literate-idris",
    "litzasic",
    "litzmax",
    "live-script",
    "livescript",
    "llvm",
    "logos",
    "logtalk",
    "lua",
    "m2",
    "make",
    "make",
    "mako",
    "maql",
    "mask",
    "mason",
    "mathematica",
    "matla",
    "matlasession",
    "mawk",
    "max",
    "md",
    "minid",
    "mma",
    "modelica",
    "modula2",
    "moin",
    "monkey",
    "monte",
    "moo",
    "moocode",
    "moon",
    "moonscript",
    "mq4",
    "mq5",
    "mql",
    "mql4",
    "mql5",
    "msc",
    "mscgen",
    "mupad",
    "mxml",
    "myghty",
    "mysql",
    "n",
    "nasm",
    "nawk",
    "ncl",
    "nesc",
    "newlisp",
    "newspeak",
    "ng2",
    "nginx",
    "nim",
    "nimrod",
    "nit",
    "nix",
    "nixos",
    "nn",
    "nsh",
    "nsi",
    "nsis",
    "numpy",
    "nusmv",
    "oa",
    "ocaml",
    "octave",
    "odin",
    "oj-c",
    "oj-c++",
    "oj-j",
    "ojc",
    "ojc++",
    "ojdump",
    "ojdump-nasm",
    "ojective-c",
    "ojective-c++",
    "ojective-j",
    "ojectivec",
    "ojectivec++",
    "ojectivej",
    "ojectpascal",
    "ojj",
    "oo",
    "ooc",
    "oogie",
    "opa",
    "openedge",
    "openugs",
    "pacmancon",
    "pan",
    "parasail",
    "pas",
    "pascal",
    "pawn",
    "pcmk",
    "perl",
    "perl6",
    "php",
    "php3",
    "php4",
    "php5",
    "pig",
    "pike",
    "pkgcon",
    "pl",
    "pl6",
    "plpgsql",
    "plus",
    "po",
    "posh",
    "postgres",
    "postgres-console",
    "postgresql",
    "postgresql-console",
    "postscr",
    "postscript",
    "pot",
    "pov",
    "powershell",
    "praat",
    "progress",
    "prolog",
    "properties",
    "protou",
    "ps1",
    "ps1con",
    "psm1",
    "psql",
    "pug",
    "puppet",
    "py",
    "py3",
    "py3t",
    "pycon",
    "pypy",
    "pypylog",
    "pyrex",
    "pyt",
    "python",
    "python3",
    "pyx",
    "qasic",
    "qml",
    "qs",
    "qvt",
    "qvto",
    "r",
    "racket",
    "ragel",
    "ragel-c",
    "ragel-cpp",
    "ragel-d",
    "ragel-em",
    "ragel-java",
    "ragel-ojc",
    "ragel-r",
    "ragel-ruy",
    "rain",
    "raw",
    "rcon",
    "rd",
    "red",
    "red/system",
    "redcode",
    "registry",
    "reol",
    "resource",
    "resourceundle",
    "rexx",
    "rhtml",
    "rkt",
    "rnc",
    "rng-compact",
    "ro",
    "roocon",
    "root",
    "rql",
    "rs",
    "rsl",
    "rts",
    "rust",
    "ruy",
    "s",
    "sage",
    "salt",
    "sarl",
    "sas",
    "sass",
    "satch",
    "sc",
    "scala",
    "scaml",
    "scheme",
    "scila",
    "scm",
    "scss",
    "shell-session",
    "silver",
    "slash",
    "slim",
    "sls",
    "slurm",
    "smali",
    "smalltalk",
    "smarty",
    "sml",
    "snool",
    "sources.list",
    "sourceslist",
    "sp",
    "sparql",
    "spec",
    "spit",
    "splus",
    "sql",
    "sqlite3",
    "squeak",
    "squidcon",
    "st",
    "st-pytex",
    "stan",
    "stata",
    "supercollider",
    "sv",
    "swi",
    "swig",
    "systemverilog",
    "t-sql",
    "tads3",
    "tap",
    "tasm",
    "tcl",
    "tcsh",
    "tcshcon",
    "tea",
    "teraterm",
    "teratermmacro",
    "termcap",
    "termin",
    "terra",
    "thri",
    "todotxt",
    "toml",
    "tra",
    "trac-wiki",
    "treetop",
    "ts",
    "tsql",
    "ttl",
    "turtle",
    "twig",
    "typescript",
    "typoscript",
    "typoscriptcssdata",
    "typoscripthtmldata",
    "ucode",
    "ugs",
    "unicon",
    "uriscript",
    "v.net",
    "vala",
    "vapi",
    "vcl",
    "vclsnippet",
    "vclsnippets",
    "vctreestatus",
    "velocity",
    "vgl",
    "vhdl",
    "vim",
    "vnet",
    "vscript",
    "wdi",
    "whiley",
    "winatch",
    "winugs",
    "x10",
    "xml",
    "xml+cheetah",
    "xml+django",
    "xml+er",
    "xml+evoque",
    "xml+genshi",
    "xml+jinja",
    "xml+kid",
    "xml+lasso",
    "xml+mako",
    "xml+myghty",
    "xml+php",
    "xml+ruy",
    "xml+smarty",
    "xml+spit",
    "xorg",
    "xq",
    "xql",
    "xqm",
    "xquery",
    "xqy",
    "xten",
    "xtend",
    "yaml",
    "yaml+jinja",
]



import os


class TLVAst(object):

    def __init__(self, filename, line_offset=0):
        self.Processed = []
        self.Raw = []
        self.Filename = filename
        self.LineOffset = line_offset


class TLVFinding(object):
    ID_DUPLICATE = "Duplicate"
    ID_TLV = "TooLessVariation"

    def __init__(self, first, second, match, _id):
        self.file_this = first.Filename
        self.file_other = second.Filename
        self.file_this_start_line, self.file_this_start_col = self.__get_pos(
            first, match.a)
        self.file_other_start_line, self.file_other_start_col = self.__get_pos(
            second, match.b)
        self.file_this_stop_line, self.file_this_stop_col = self.__get_pos(
            first, match.a + match.size)
        self.file_other_stop_line, self.file_other_stop_col = self.__get_pos(
            second, match.b + match.size)
        self.text_this = TLVFinding.get_text(
            first, match.a, match.a + match.size)
        self.text_other = TLVFinding.get_text(
            second, match.b, match.b + match.size)
        self.id = _id

    def Print(self, args):
        print(str(self))
        self.PrintDetails(args)

    def PrintDetails(self, args):
        if args.details:
            print(">>> {snip}".format(snip=self.text_this.replace(
                os.linesep, os.linesep + ">>> ")))
            print("<<<")
            if self.id == TLVFinding.ID_TLV:
                print(">>> {snip}".format(snip=self.text_other.replace(
                    os.linesep, os.linesep + ">>> ")))
                print("<<<")

    def __eq__(self, other):
        if isinstance(other, TLVFinding):
            return str(other) == str(self)
        else:
            return False

    def __ne__(self, other):
        return (not self.__eq__(other))

    def __hash__(self):
        return hash(self.__repr__())

    def __repr__(self):
        if self.id == TLVFinding.ID_DUPLICATE:
            msg = "is the same"
        elif self.id == TLVFinding.ID_TLV:
            msg = "is nearly the same"
        if self.file_this != self.file_other:
            path = os.path.relpath(self.file_other, self.file_this)
        else:
            path = "same file"
        return "{}:{}:{}:[{}]:Block till {}:{} {} as in {} from {}:{} till {}:{}".format(self.file_this,  # noqa: P101
                                                                                         self.file_this_start_line,
                                                                                         self.file_this_start_col,
                                                                                         self.id,
                                                                                         self.file_this_stop_line,
                                                                                         self.file_this_stop_col,
                                                                                         msg,
                                                                                         path,
                                                                                         self.file_other_start_line,
                                                                                         self.file_other_start_col,
                                                                                         self.file_other_stop_line,
                                                                                         self.file_other_stop_col)

    def __get_pos(self, obj, pos):
        _raw = TLVFinding.get_text(obj, 0, pos)
        _line = _raw.count(os.linesep)
        _col = len(_raw.split(os.linesep)[-1])
        return (_line + obj.LineOffset, _col)

    @staticmethod
    def get_text(obj, start, stop):
        return "".join([x[1] for x in obj.Raw[start:stop]])

    @staticmethod
    def Validate(args, first, second, match):
        chunk_a = TLVFinding.get_text(first, match.a, match.a + match.size)
        chunk_b = TLVFinding.get_text(second, match.b, match.b + match.size)
        if chunk_a.count(os.linesep) + 1 >= args.minlines and match.size >= args.mintoken:
            if chunk_a == chunk_b:
                return TLVFinding.ID_DUPLICATE
            else:
                return TLVFinding.ID_TLV
        return None



TOKENS = [
    "Token.Keyword",
    "Token.Keyword.Constant",
    "Token.Keyword.Declaration",
    "Token.Keyword.Namespace",
    "Token.Keyword.Pseudo",
    "Token.Keyword.Reserved",
    "Token.Keyword.Type",
    "Token.Name",
    "Token.Name.Attribute",
    "Token.Name.Builtin",
    "Token.Name.Builtin.Pseudo",
    "Token.Name.Class",
    "Token.Name.Constant",
    "Token.Name.Decorator",
    "Token.Name.Entity",
    "Token.Name.Exception",
    "Token.Name.Function",
    "Token.Name.Function.Magic",
    "Token.Name.Label",
    "Token.Name.Namespace",
    "Token.Name.Other",
    "Token.Name.Tag",
    "Token.Name.Variable",
    "Token.Name.Variable.Class",
    "Token.Name.Variable.Global",
    "Token.Name.Variable.Instance",
    "Token.Name.Variable.Magic",
    "Token.Literal",
    "Token.Literal.Date",
    "Token.Literal.String",
    "Token.Literal.String.Affix",
    "Token.Literal.String.Backtick",
    "Token.Literal.String.Char",
    "Token.Literal.String.Delimiter",
    "Token.Literal.String.Doc",
    "Token.Literal.String.Double",
    "Token.Literal.String.Escape",
    "Token.Literal.String.Heredoc",
    "Token.Literal.String.Interpol",
    "Token.Literal.String.Other",
    "Token.Literal.String.Regex",
    "Token.Literal.String.Single",
    "Token.Literal.String.Symbol",
    "Token.Literal.Number",
    "Token.Literal.Number.Bin",
    "Token.Literal.Number.Float",
    "Token.Literal.Number.Hex",
    "Token.Literal.Number.Integer",
    "Token.Literal.Number.Integer.Long",
    "Token.Literal.Number.Oct",
    "Token.Operator",
    "Token.Operator.Word",
    "Token.Punctuation",
    "Token.Comment",
    "Token.Comment.Hashbang",
    "Token.Comment.Multiline",
    "Token.Comment.Preproc",
    "Token.Comment.Single",
    "Token.Comment.Special",
    "Token.Generic",
    "Token.Generic.Deleted",
    "Token.Generic.Emph",
    "Token.Generic.Error",
    "Token.Generic.Heading",
    "Token.Generic.Inserted",
    "Token.Generic.Output",
    "Token.Generic.Prompt",
    "Token.Generic.Strong",
    "Token.Generic.Subheading",
    "Token.Generic.Traceback",
    "Token.Text.Whitespace",
]



import sys


def debug(args, msg):
    if args.verbose:
        print(msg)


def die(code, msg=None):
    if msg:
        if not msg.endswith("\n"):
            msg += "\n"
        sys.stderr.write(msg)
    sys.exit(code)



import setuptools

with open('README.md') as f:
    _long_description = f.read()

requirements = []
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="tlv",
    version="1.5.1",
    author="Konrad Weihmann",
    author_email="kweihmann@outlook.com",
    description="-- Too less variation -- Find duplicates in source code for various languages",
    long_description=_long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/priv-kweihmann/tlv",
    packages=setuptools.find_packages(),
    install_requires=requirements,
        entry_points={
        "console_scripts": [
            "tlv = tlv.__main__:main",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Quality Assurance",
    ],
    python_requires='>=3.8',
)



import argparse
import os

from tlv.lexer import LEXERS
from tlv.tokens import TOKENS

RUNARGS = None


def ArgParser():
    parser = argparse.ArgumentParser(
        prog="tlv", description='Find code duplications across various languages')
    parser.add_argument("--minlines", type=int, default=5,
                        help="Minimum length of block in lines before reporting")
    parser.add_argument("--mintoken", type=int, default=20,
                        help="Minimum length of block in token before reporting")
    parser.add_argument("-j", "--jobs", type=int, default=1,
                        help="Number of jobs to run in parallel")
    parser.add_argument("--lexer", choices=LEXERS, default=None,
                        help="Manually set a lexer to use on all files")
    parser.add_argument("--wildcard_token", choices=TOKENS, action='append',
                        default=[
                            "Token.Comment.Multiline",
                            "Token.Comment.Preproc",
                            "Token.Literal.Number",
                            "Token.Literal.Number.Bin",
                            "Token.Literal.Number.Float",
                            "Token.Literal.Number.Hex",
                            "Token.Literal.Number.Integer",
                            "Token.Literal.Number.Integer.Long",
                            "Token.Literal.Number.Oct",
                            "Token.Text",
                            "Token.Whitespace",
                        ],
                        help="Token types that are threated as wildcards (actual value doesn't matter)")
    parser.add_argument("--verbose", default=False,
                        action="store_true", help="Verbose output")
    parser.add_argument("--nodetails", default=True, action="store_false",
                        dest="details", help="Dump the details of a finding")
    parser.add_argument("files", nargs='+', help="Files to parse")
    RUNARGS = parser.parse_args()
    # Turn all paths to abs-paths right here
    RUNARGS.files = [os.path.abspath(x) for x in RUNARGS.files]
    return RUNARGS
